/*
** Copyright (C) Red Hat, Inc. 2005-2008
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "Socket.h"
#include "Logger.h"

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include "String.h"

ServerSocket::ServerSocket(const String& sock_path) :
	Socket(-1),
	_unix_sock(true),
	_sock_path(sock_path)
{
	_sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (_sock == -1) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): socket() failed: " + String(strerror(errno));
	}

	int t = 1;
	if (setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, &t, sizeof(t))) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): set SO_REUSEADDR, failed: " + String(strerror(errno));
	}

	struct sockaddr_un addr;
	addr.sun_family = AF_UNIX;
	memcpy(addr.sun_path, sock_path.c_str(), sock_path.size() + 1);

	unlink(_sock_path.c_str());

	if (bind(_sock, (struct sockaddr*) &addr, sizeof(addr))) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): bind() failed: " + String(strerror(errno));
	}

	if (listen(_sock, 5)) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): listen() failed: " + String(strerror(errno));
	}
	//String msg = String("created unix server socket, ")
	//				+ _sock + ", " + sock_path;
	//log(msg, LogSocket);
}

ServerSocket::ServerSocket(unsigned short port) :
	Socket(-1),
	_unix_sock(false),
	_sock_path("")
{
	_sock = socket(PF_INET, SOCK_STREAM, 0);
	if (_sock == -1) {
		throw String("ServerSocket(port=") + port
				+ "): socket() failed: " + String(strerror(errno));
	}

	int t = 1;
	if (setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, &t, sizeof(t))) {
		throw String("ServerSocket(port=") + port
				+ "): set SO_REUSEADDR, failed: " + String(strerror(errno));
	}

	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;

	if (bind(_sock, (struct sockaddr*) &addr, sizeof(addr))) {
		throw String("ServerSocket(port=") + port
				+ "): bind() failed: " + String(strerror(errno));
	}

	if (listen(_sock, 5)) {
		throw String("ServerSocket(port=") + port
				+ "): listen() failed: " + String(strerror(errno));
	}

	//String msg = String("created tcp server socket, ")
	//				+ _sock + ", port " + port;
	//log(msg, LogSocket);
}

ServerSocket::ServerSocket(const ServerSocket& s) :
	Socket(s),
	_unix_sock(s._unix_sock),
	_sock_path(s._sock_path)
{}

ServerSocket&
ServerSocket::operator= (const ServerSocket& s)
{
	if (&s != this) {
		this->Socket::operator= (s);
		_unix_sock = s._unix_sock;
		_sock_path = s._sock_path;
	}

	return *this;
}

ServerSocket::~ServerSocket()
{
	if (_unix_sock && (*_counter == 1))
		unlink(_sock_path.c_str());
}

ClientSocket
ServerSocket::accept()
{
	while (true) {
		struct sockaddr_in addr_in;
		socklen_t size = sizeof(addr_in);

		int ret = ::accept(_sock, (struct sockaddr *) &addr_in, &size);
		if (ret == -1) {
			if (errno == EINTR)
				continue;
			throw String("ServerSocket(): accept() failed: ")
					+ String(strerror(errno));
		}
		//log("ServerSocket: accepted connection", LogSocket);
		return ClientSocket(ret, addr_in.sin_addr.s_addr);
	}
}

bool
ServerSocket::ready(int timeout)
{
	bool read = true;
	bool write = false;

	poll(read, write, timeout);
	return read;
}
