/*
** Copyright (C) Red Hat, Inc. 2006-2008
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "PackageHandler.h"
#include "utils.h"
#include "File.h"

#include <unistd.h>
#include <sys/utsname.h>

using namespace std;

#define RPM_PATH		"/bin/rpm"
#define UP2DATE_PATH	"/usr/sbin/up2date-nox"
#define YUM_PATH		"/usr/bin/yum"

static bool RHEL4=false, RHEL5=false, FC5=false, FC6=false;
static void set_release();

// class PackageInstaller

PackageInstaller::PackageInstaller()
{
	set_release();
}

PackageInstaller::~PackageInstaller()
{}

bool
PackageInstaller::available()
{
	if (RHEL4) {
		bool rhn_available = true;
		// check existence of /etc/sysconfig/rhn/systemid
		if (access("/etc/sysconfig/rhn/systemid", F_OK))
			rhn_available = false;

		// check for subscribed channels
		if (rhn_available) {
			String out, err;
			int status;
			vector<String> args;

			args.push_back("--show-channels");
			if (utils::execute(UP2DATE_PATH, args, out, err, status, false))
				throw command_not_found_error_msg(UP2DATE_PATH);
			if (status != 0)
				rhn_available = false;
		}
		return rhn_available;
	} else if (FC5 || FC6 || RHEL5) {
		// use yum
		// nothing to check for, maybe ping repositories?
		return true;
	} else
		return false;
}

map<String, String>
PackageInstaller::available_rpms()
{
	map<String, String> rpms;

	if (!available())
		return rpms;

	if (RHEL4) {
		String out, err;
		int status;
		vector<String> args;

		args.push_back("--showall");
		if (utils::execute(UP2DATE_PATH, args, out, err, status, false))
			throw command_not_found_error_msg(UP2DATE_PATH);
		vector<String> lines = utils::split(utils::strip(out), "\n");

		for (vector<String>::const_iterator
				iter = lines.begin() ;
				iter != lines.end() ;
				iter++)
		{
			String line(*iter);
			line = utils::strip(line);
			vector<String> words = utils::split(line, "-");
			vector<String>::size_type l = words.size();
			if (l < 3)
				continue;

			String name = words[0];
			for (unsigned int i = 1 ; i < l - 2 ; i++)
				name += "-" + words[i];
			String version = words[l - 2] + "-" + words[l - 1].substr(0, words[l - 1].rfind("."));
			rpms[name] = version;
		}
	} else if (FC5 || FC6 || RHEL5) {
		String out, err;
		int status;
		vector<String> args;

		args.push_back("-y");
		args.push_back("list");
		args.push_back("all");

		if (utils::execute(YUM_PATH, args, out, err, status, false))
			throw command_not_found_error_msg(YUM_PATH);

		vector<String> lines = utils::split(utils::strip(out), "\n");

		for (vector<String>::const_iterator
				iter = lines.begin() ;
				iter != lines.end() ;
				iter++)
		{
			String line(*iter);
			line = utils::strip(line);
			vector<String> words = utils::split(line);
			vector<String>::size_type l = words.size();
			if (l < 3)
				continue;

			String name = words[0];
			String::size_type idx = name.rfind('.');
			if (idx == String::npos)
				continue;
			name = name.substr(0, idx);
			String version = words[1];
			rpms[name] = version;
		}
	}

	return rpms;
}

bool
PackageInstaller::install(vector<String> rpms)
{
	if (rpms.empty())
		return true;

	if (RHEL4) {
		String out, err;
		int status;
		vector<String> args;

		for (vector<String>::size_type i = 0 ; i < rpms.size() ; i++)
			args.push_back(rpms[i]);

		if (utils::execute(UP2DATE_PATH, args, out, err, status, true))
			throw command_not_found_error_msg(UP2DATE_PATH);

		if (status != 0)
			throw String(UP2DATE_PATH) + " failed: " + err;
		return true;
	} else if (FC5 || FC6 || RHEL5) {
		vector<String> rpms_to_install, rpms_to_upgrade;
		String out, err;
		int status;
		vector<String> args;

		args.push_back("-y");
		args.push_back("list");
		args.push_back("installed");

		if (utils::execute(YUM_PATH, args, out, err, status, false))
			throw command_not_found_error_msg(YUM_PATH);
		if (status)
			return false;

		vector<String> lines = utils::split(utils::strip(out), "\n");
		for (vector<String>::const_iterator
				rpm = rpms.begin() ;
				rpm != rpms.end() ;
				rpm++)
		{
			bool install = true;
			for (vector<String>::const_iterator
					iter = lines.begin() ;
					iter != lines.end() ;
					iter++)
			{
				String line(*iter);
				line = utils::strip(line);
				if (line.find(*rpm + ".") == 0)
					install = false;
			}

			if (install)
				rpms_to_install.push_back(*rpm);
			else
				rpms_to_upgrade.push_back(*rpm);
		}

		if (!rpms_to_install.empty()) {
			out = err = "";
			args.clear();

			args.push_back("-y");
			args.push_back("install");
			for (vector<String>::const_iterator
					rpm = rpms_to_install.begin() ;
					rpm != rpms_to_install.end() ;
					rpm++)
			{
				args.push_back(*rpm);
			}

			if (utils::execute(YUM_PATH, args, out, err, status, false))
				throw command_not_found_error_msg(YUM_PATH);
			if (status)
				return false;
		}

		if (!rpms_to_upgrade.empty()) {
			out = err = "";
			args.clear();

			args.push_back("-y");
			args.push_back("update");
			for (vector<String>::const_iterator
					rpm = rpms_to_upgrade.begin() ;
					rpm != rpms_to_upgrade.end() ;
					rpm++)
			{
				args.push_back(*rpm);
			}

			if (utils::execute(YUM_PATH, args, out, err, status, false))
				throw command_not_found_error_msg(YUM_PATH);
			if (status)
				return false;
		}
		return true;
	} else
		return false;
}

// #### class Package ####

Package::Package()
{
	set_release();
}

Package::Package(const String& name) :
	name(name)
{
	set_release();
}

Package::~Package()
{}

XMLObject
Package::xml() const
{
	XMLObject xml("rpm");
	xml.set_attr("name", name);
	xml.set_attr("summary", summary);
	xml.set_attr("description", description);
	xml.set_attr("version", version);
	xml.set_attr("repo_version", repo_version);
	return xml;
}

// #### class PackageSet ####

PackageSet::PackageSet() :
	installed(false),
	in_repo(false),
	upgradeable(false)
{
	set_release();
}

PackageSet::PackageSet(const String& name) :
	name(name),
	installed(false),
	in_repo(false),
	upgradeable(false)
{
	set_release();
}

PackageSet::~PackageSet()
{}

XMLObject
PackageSet::xml() const
{
	XMLObject xml("set");
	xml.set_attr("name", name);
	xml.set_attr("summary", summary);
	xml.set_attr("description", description);
	xml.set_attr("installed", installed ? "true" : "false");
	xml.set_attr("in_repository", in_repo ? "true" : "false");
	xml.set_attr("installable", upgradeable ? "true" : "false");
	return xml;
}

// #### class PackageHandler ####

PackageInstaller PackageHandler::_pi;

PackageHandler::PackageHandler()
{
	set_release();

	// get installed packages
	String out, err;
	int status;
	vector<String> args;

	args.push_back("-qa");
	if (utils::execute(RPM_PATH, args, out, err, status, false))
		throw command_not_found_error_msg(RPM_PATH);
	if (status != 0)
		throw String("rpm -qa failed: " + err);

	vector<String> lines = utils::split(out, "\n");
	for (vector<String>::const_iterator
			iter = lines.begin() ;
			iter != lines.end() ;
			iter++)
	{
		String line(*iter);
		line = utils::strip(line);

		vector<String> words = utils::split(line, "-");
		vector<String>::size_type l = words.size();
		if (l < 3)
			continue;
		String name = words[0];

		for (unsigned int i = 1; i < l - 2 ; i++)
			name += "-" + words[i];

		Package pack(name);
		pack.version = words[l - 2] + "-" + words[l - 1];
		_packages[name] = pack;
	}

	// probe repositories
	if (repo_available()) {
		map<String, String> avail_rpms = _pi.available_rpms();
		for (map<String, String>::const_iterator
			iter = avail_rpms.begin() ;
			iter != avail_rpms.end() ;
			iter++)
		{
			String name = iter->first;
			String version = iter->second;
			Package& pack = _packages[name];
			pack.name = name;
			pack.repo_version = version;
		}
	}

	// build sets
	_sets = build_sets();
	for (map<String, PackageSet>::iterator
			iter = _sets.begin() ;
			iter != _sets.end() ;
			iter++)
	{
		populate_set(iter->second);
	}
}

PackageHandler::~PackageHandler()
{}

std::map<String, Package>&
PackageHandler::packages()
{
	return _packages;
}

std::map<String, PackageSet>&
PackageHandler::sets()
{
	return _sets;
}

std::map<String, PackageSet>
PackageHandler::build_sets()
{
	map<String, PackageSet> sets;

	PackageSet set = build_cluster_base_set();
	sets[set.name] = set;

	if (RHEL4 || FC5) {
		set = build_cluster_base_gulm_set();
		sets[set.name] = set;
	}

	set = build_cluster_services_set();
	sets[set.name] = set;

	set = build_cluster_storage_set();
	sets[set.name] = set;

	set = build_linux_virtual_server_set();
	sets[set.name] = set;

	return sets;
}

PackageSet
PackageHandler::build_cluster_base_set()
{
	PackageSet set("Cluster Base");

	if (RHEL4 || FC5) {
		set.packages.push_back("ccs");
		set.packages.push_back("cman");
		set.packages.push_back("dlm");
		set.packages.push_back("fence");

		struct utsname uts;
		uname(&uts);
		String kernel(uts.release);

		if (kernel.find("smp") != kernel.npos) {
			set.packages.push_back("cman-kernel-smp");
			set.packages.push_back("dlm-kernel-smp");
		} else if (kernel.find("hugemem") != kernel.npos) {
			set.packages.push_back("cman-kernel-hugemen");
			set.packages.push_back("dlm-kernel-hugemem");
		} else {
			set.packages.push_back("cman-kernel");
			set.packages.push_back("dlm-kernel");
		}
	} else if (RHEL5 || FC6) {
		set.packages.push_back("cman");
	}

	return set;
}

PackageSet
PackageHandler::build_cluster_base_gulm_set()
{
	PackageSet set("Cluster Base - Gulm");

	if (RHEL4 || FC5) {
		set.packages.push_back("ccs");
		set.packages.push_back("gulm");
	} else
		throw String("GULM is not available in Red Hat Cluster Suite 5");

	return set;
}

PackageSet
PackageHandler::build_cluster_services_set()
{
	PackageSet set("Cluster Service Manager");
	set.packages.push_back("rgmanager");
	if (RHEL4 || FC5) {
		set.packages.push_back("magma");
		set.packages.push_back("magma-plugins");
	}
	return set;
}

PackageSet
PackageHandler::build_cluster_storage_set()
{
	struct utsname uts;
	uname(&uts);
	String kernel(uts.release);

	PackageSet set("Clustered Storage");
	set.packages.push_back("lvm2-cluster");
	set.packages.push_back("sg3_utils");

	if (RHEL4 || FC5) {
		set.packages.push_back("GFS");
		if (kernel.find("smp") != kernel.npos) {
			set.packages.push_back("GFS-kernel-smp");
		} else if (kernel.find("hugemem") != kernel.npos) {
			set.packages.push_back("GFS-kernel-hugemem");
		} else {
			set.packages.push_back("GFS-kernel");
		}
	} else if (RHEL5 || FC6) {
		set.packages.push_back("gfs2-utils");
		set.packages.push_back("cmirror");
		if (RHEL5) {
			set.packages.push_back("gfs-utils");
			if (kernel.find("xen") == kernel.npos) {
				set.packages.push_back("kmod-gfs");
				set.packages.push_back("kmod-cmirror");
			} else {
				set.packages.push_back("kmod-gfs-xen");
				set.packages.push_back("kmod-cmirror-xen");
			}
		}
	}

	return set;
}

PackageSet
PackageHandler::build_linux_virtual_server_set()
{
	PackageSet set("Linux Virtual Server");
	set.packages.push_back("ipvsadm");
	set.packages.push_back("piranha");
	return set;
}

void
PackageHandler::populate_set(PackageSet& set)
{
	set.installed = true;
	set.in_repo = true;
	set.upgradeable = false;

	for (list<String>::const_iterator
			name_iter = set.packages.begin() ;
			name_iter != set.packages.end() ;
			name_iter++)
	{
		const String& name = *name_iter;
		map<String, Package>::const_iterator iter = _packages.find(name);
		if (iter == _packages.end()) {
			set.installed = false;
			set.in_repo = false;
			set.upgradeable = false;
			break;
		} else {
			const Package& pack = iter->second;
			if (pack.version.empty())
				set.installed = false;
			if (pack.repo_version.empty())
				set.in_repo = false;
			else if (pack.repo_version > pack.version)
				set.upgradeable = true;
		}
	}

	if (set.in_repo == false)
		set.upgradeable = false;
}

void
PackageHandler::install(const std::list<Package>& packages,
						const std::list<PackageSet>& sets,
						bool upgrade)
{
	vector<String> rpms;

	PackageHandler h_pre;
	for (list<Package>::const_iterator
			iter = packages.begin() ;
			iter != packages.end() ;
			iter++)
	{
		String name(iter->name);
		map<String, Package>::iterator pack_iter = h_pre.packages().find(name);
		if (pack_iter == h_pre.packages().end()) {
			throw String("Package \"") + name
					+ "\" is present neither locally nor in any available repository";
		} else {
			String curr_ver(pack_iter->second.version);
			String repo_ver(pack_iter->second.repo_version);
			if (curr_ver.empty()) {
				// not installed
				if (repo_ver.empty()) {
					throw String("Package \"") + name
							+ "\" is not present in any available repository";
				} else
					rpms.push_back(name);
			} else {
				// already installed
				if (upgrade) {
					if (repo_ver.empty()) {
						throw String("Package \"") + name
								+ "\" is not present in any available repository";
					} else if (repo_ver > curr_ver)
						rpms.push_back(name);
				}
			}
		}
	}

	for (list<PackageSet>::const_iterator
			iter = sets.begin() ;
			iter != sets.end() ;
			iter++)
	{
		String name(iter->name);
		map<String, PackageSet>::iterator set_iter = h_pre.sets().find(name);

		if (set_iter == h_pre.sets().end()) {
			throw String("Packages of set \"") + name
					+ "\" are neither present neither locally nor in any available repository";
		} else {
			PackageSet& p_set = set_iter->second;
			if (p_set.installed) {
				// already installed
				if (upgrade) {
					if (p_set.in_repo) {
						if (p_set.upgradeable) {
							for (list<String>::const_iterator
									name_iter = p_set.packages.begin() ;
									name_iter != p_set.packages.end() ;
									name_iter++)
							{
								rpms.push_back(*name_iter);
							}
						} else {
							/* Packages are already up-to-date */
						}
					}
				}
			} else {
				// not installed
				if (p_set.in_repo) {
					for (list<String>::const_iterator
							name_iter = p_set.packages.begin() ;
							name_iter != p_set.packages.end() ;
							name_iter++)
					{
						rpms.push_back(*name_iter);
					}
				} else {
					throw String("Packages of set \"") + name +
						"\" are not present in any available repository";
				}
			}
		}
	}

	if (!_pi.install(rpms)) {
		String msg("Failed to install packages");
		if (!repo_available())
			msg += ": System not configured to use repositories";
		throw msg;
	}
}

bool
PackageHandler::repo_available()
{
	return _pi.available();
}

void
set_release()
{
	static bool release_set = false;
	if (!release_set) {
		String release(utils::strip(File::open("/etc/redhat-release")));

		if (release.find("Nahant") != release.npos)
			RHEL4 = true;
		else if (release.find("Bordeaux") != release.npos)
			FC5 = true;
		else if (release.find("Zod") != release.npos)
			FC6 = true;
		else if (release.find("Moonshine") != release.npos)
			FC6 = true;
		else if (release.find("Werewolf") != release.npos)
			FC6 = true;
		else if (release.find("Sulphur") != release.npos)
			FC6 = true;
		else if (release.find("Tikanga") != release.npos)
			RHEL5 = true;
		else
			throw String("Unsupported OS: " + release);

		release_set = true;
	}
}
